@@ -5,16 +5,17 @@ use base qw/DBIx::Class::Schema/;
 
 no warnings qw/qw/;
 
+__PACKAGE__->mk_group_accessors(simple => 'custom_attr');
+
 __PACKAGE__->load_classes(qw/
   Artist
   SequenceTest
   BindType
   Employee
   CD
-  FileColumn
   Genre
-  Link
   Bookmark
+  Link
   #dummy
   Track
   Tag
